from sly import Parser
from .swas_lexer import SwasLexer

class SwasParser(Parser):
    tokens = SwasLexer.tokens
    #debugfile = "log.out"
    precedence = (
        ('left', OR, AND), 
        ('left', EQ, LT , GT ,NE, GTE, LTE), 
        ('left', PLUS, MINUS),
        ('left', TIMES, DIVIDE),   
        ('left', POW), 
        ('left', MOD),
        ('right', UMINUS),
        ('left', WHILE, DO),
        ('left', IF, ELIF, ELSE),  
        ('left', PRINT, INPUT)
        )

    def __init__(self):
        self.names = { }
        self.prompt = True

    ############################################################
    # MAIN
    ############################################################

    @_('statements')
    def main(self, p):
        return ('main', p.statements)

    @_('statement')
    def statements(self,p):
        return ('statements', [p.statement])

    @_('statements statement')
    def statements(self,p):
        return ('statements', p.statements[1] + [p.statement])


    ############################################################
    # STATEMENTS
    ############################################################

    @_('PRINT statement')
    def statement(self, p):
        return ('print', p.statement)

    @_('INPUT statement')
    def statement(self, p):
        return ('input', p.statement)

    @_('expr')
    def statement(self, p):
        return ('statement-expr', p.expr)

    @_('NAME ARROW statement')
    def statement(self, p):
        return ('assign', p.NAME, p.statement)

    @_('IF expr LBRAC statements RBRAC [ ELIF expr LBRAC statements RBRAC ] [ ELSE LBRAC statements RBRAC ] ')
    def statement(self, p):
        return ('if-elif-else', p.expr0 ,p.statements0, p.expr1, p.statements1, p.statements2)

    @_('WHILE expr DO LBRAC statements RBRAC')
    def statement(self, p):
        return ('while', p.expr, p.statements)

    @_('PASS')
    def statement(self, p):
        return ('pass',)

    @_('BREAK')
    def statement(self, p):
        return ('break',)

    ############################################################
    # Expressions
    ############################################################


    @_('expr PLUS expr')
    def expr(self, p):
        return ('plus', p.expr0, p.expr1)

    @_('expr MINUS expr')
    def expr(self, p):
        return ('minus', p.expr0, p.expr1)

    @_('expr TIMES expr')
    def expr(self, p):
        return ('times', p.expr0, p.expr1)

    @_('expr DIVIDE expr')
    def expr(self, p):
        return ('divide', p.expr0, p.expr1)

    @_('expr MOD expr')
    def expr(self, p):
        return ('mod', p.expr0, p.expr1)

    @_('expr POW expr')
    def expr(self, p):
        return ('pow', p.expr0, p.expr1)
   
    @_('expr EQ expr')
    def expr(self, p):
        return ('equals', p.expr0, p.expr1)

    @_('expr NE expr')
    def expr(self, p):
        return ('ne', p.expr0, p.expr1)

    @_('expr GT expr')
    def expr(self, p):
        return ('gt', p.expr0, p.expr1)

    @_('expr GTE expr')
    def expr(self, p):
        return ('gte', p.expr0, p.expr1)

    @_('expr LT expr')
    def expr(self, p):
        return ('lt', p.expr0, p.expr1)

    @_('expr LTE expr')
    def expr(self, p):
        return ('lte', p.expr0, p.expr1)

    @_('expr AND expr')
    def expr(self, p):
        return ('and', p.expr0, p.expr1)

    @_('expr OR expr')
    def expr(self, p):
        return ('or', p.expr0, p.expr1)


    @_('INC NAME %prec UMINUS')
    def expr(self, p):
        return ('inc', p.NAME)    

    @_('DEC NAME %prec UMINUS')
    def expr(self, p):
        return ('dec', p.NAME)    

    # Weird
    # @_('MINUS expr %prec UMINUS')
    # def expr(self, p):
    #     return ('uminus', p.expr)

    @_('LPAREN expr RPAREN')
    def expr(self, p):
        return ('paren', p.expr)

    @_('NUMBER')
    def expr(self, p):
        return ('number', p.NUMBER)

    @_('STRING')
    def expr(self, p):
        return ('string', p.STRING)

    @_('NAME')
    def expr(self, p):
        return ('name', p.NAME)



from .swas_lexer import SwasLexer
from .swas_parser import SwasParser

VERSION = "1.8"

names = {}

def get_obj_type(_obj):
    return str(type(_obj)).split("'")[1]

def evaluate(tree):
    global names
    undefined = "Swas says: {0} isn't been defined!"
    type_error = "Swas says: You can't use '{op}' with types '{obj1}' and '{obj2}'!"
    single_te = "Swas says: You can't use '{op}' with type '{obj}'!"

    try:
        rule = tree[0]
    except TypeError:
        return print("Swas says: Error!!")

    if rule == 'main':
        evaluate(tree[1])
    elif rule == 'statements':
        results = []
        for i in tree[1]:
            results.append(evaluate(i))
        return results
    elif rule == 'statement-expr':
        value = evaluate(tree[1])
        return value
    elif rule == 'assign':
        value = evaluate(tree[2])
        name = tree[1]
        names[name] = value
        return value

    elif rule == 'times':
        multiplier = evaluate(tree[1])
        multiplicand = evaluate(tree[2])
        try:
            return multiplier * multiplicand
        except TypeError:
            return print(type_error.format(op="*", obj1=get_obj_type(multiplier), obj2=get_obj_type(multiplicand)))
    elif rule == 'plus':
        addend1 = evaluate(tree[1])
        addend2 = evaluate(tree[2])
        try:
            return addend1 + addend2
        except TypeError:
            return print(type_error.format(op="+", obj1=get_obj_type(addend1), obj2=get_obj_type(addend2)))
    elif rule == 'minus':
        minuend = evaluate(tree[1])
        subtrahend = evaluate(tree[2])
        try:
            return minuend - subtrahend
        except TypeError:
            return print(type_error.format(op="-", obj1=get_obj_type(minuend), obj2=get_obj_type(subtrahend)))
    elif rule == 'divide':
        dividend = evaluate(tree[1])
        divisor = evaluate(tree[2])
        try:
            return dividend / divisor
        except TypeError:
            return print(type_error.format(op="/", obj1=get_obj_type(dividend), obj2=get_obj_type(divisor)))
    elif rule == 'mod':
        a = evaluate(tree[1])
        n = evaluate(tree[2])
        try:
            return evaluate(a % n)
        except TypeError:
            return print(type_error.format(op="%", obj1=get_obj_type(a), obj2=get_obj_type(n)))
    elif rule == 'pow':
        target_num = tree[1]
        exponent = tree[2]
        try:
            return target_num ** exponent
        except TypeError:
            return print(type_error.format(op="^", obj1=get_obj_type(target_num), obj2=get_obj_type(exponent)))
            
    elif rule == 'equals':
        return int(evaluate(tree[1]) == evaluate(tree[2]))
    elif rule == 'ne':
        return int(evaluate(tree[1]) != evaluate(tree[2]))
    elif rule == 'gt':
        return int(evaluate(tree[1]) > evaluate(tree[2]))
    elif rule == 'gte':
        return int(evaluate(tree[1]) >= evaluate(tree[2]))
    elif rule == 'lt':
        return int(evaluate(tree[1]) < evaluate(tree[2]))
    elif rule == 'lte':
        return int(evaluate(tree[1]) <= evaluate(tree[2]))
    elif rule == 'and':
        return int(evaluate(tree[1]) and evaluate(tree[2]))
    elif rule == 'or':
        return int(evaluate(tree[1]) or evaluate(tree[2]))


    elif rule == 'uminus':
        return -evaluate(tree[1])
    elif rule == 'inc':
        name = tree[1]

        try:        
            oldval = names[tree[1]]
        except KeyError:
            return print(f"Swas says: {name} hasn't been defined!")        
        newval = oldval + 1
        
        names[name] = newval
        return newval
    elif rule == 'dec':
        name = tree[1]
        try:        
            oldval = names[tree[1]]
        except KeyError:
            return print(f"Swas says: {name} hasn't been defined!") 

        newval = oldval - 1
        
        names[name] = newval
        return newval
    elif rule == 'number':
        try:
            return int(tree[1])
        except ValueError:
            return float(tree[1])
    elif rule == 'string':
        return str(tree[1])
    elif rule == 'name':
        varname = tree[1]
        try:
            return names[varname]
        except KeyError:
            print(f"Swas says: {varname} hasn't been defined!")
    elif rule == 'paren':
        return evaluate(tree[1])
    elif rule == 'pass':
        pass
    elif rule == 'break':
        return Break()
    elif rule == 'print':
        value = evaluate(tree[1])
        print(value)
        return value
    elif rule == 'input':
        value = evaluate(tree[1])
        res = input(value)
        try:
            res = float(res)
        except ValueError:
            pass
        return res
    elif rule == 'if-elif-else':
        expr1 = evaluate(tree[1])
        expr2 = None if tree[3] is None else evaluate(tree[3])
        if expr1:
            return evaluate(tree[2])
        elif expr2:
            return evaluate(tree[4])
        else:
            if tree[5]:
                return evaluate(tree[5])
            else:
                pass
    elif rule == 'while':
        while evaluate(tree[1]):
            results = evaluate(tree[2])
            
            if any([isinstance(res,Break) for res in results]):
                break
    else:
        print(rule)

class Break:
    pass

def execute(fp):
    with open(fp, "r") as f:
        text = f.read()

    lexer = SwasLexer()
    parser = SwasParser()
    tree = parser.parse(lexer.tokenize(text))
    evaluate(tree)

def shell():
    lexer = SwasLexer()
    parser = SwasParser()
    print(f"Swas {VERSION}")
    while True:
        try:
            text = input('swas > ')
        except EOFError:
            break
        tree = parser.parse(lexer.tokenize(text))
        evaluate(tree)




from sys import argv
from .executor import execute, shell

def main():
    
    if len(argv) <= 1:
        shell()
    else:
        fp = argv[1]
        execute(fp)

if __name__ == '__main__':
    main()




from sly import Lexer

class SwasLexer(Lexer):
    tokens = { NAME, NUMBER, PLUS, TIMES, MINUS, DIVIDE, MOD, POW, ARROW, LPAREN, RPAREN,
               IF, ELIF, ELSE, WHILE, DO, BREAK, STRING, PRINT, INPUT, INC, DEC, EQ, GT, GTE, LT, LTE, NE, PASS,
               LBRAC, RBRAC, OR, AND}
    
    # Ignored patterns
    ignore_newline = r'\n+'
    ignore_comment = r'//.*\n*'
    ignore = ' \t'
    
    # Tokens
    NAME = r'[a-zA-Z_][a-zA-Z0-9_]*'
    
    NUMBER = r'\d+\.{0,1}\d*'

    NAME['if'] = IF 
    NAME['elif'] = ELIF
    NAME['else'] = ELSE
    NAME['while'] = WHILE
    NAME['do'] = DO
    NAME['break'] = BREAK
    NAME['output'] = PRINT
    NAME['input'] = INPUT
    NAME['inc'] = INC
    NAME['dec'] = DEC
    NAME['nothing'] = PASS
    NAME['and'] = AND
    NAME['or'] = OR

    # Operators
    PLUS = r'\+'
    MINUS = r'-'
    TIMES = r'\*'
    DIVIDE = r'/'
    LPAREN = r'\('
    RPAREN = r'\)'
    LBRAC = r'\{'
    RBRAC = r'\}'
    MOD = r'%'
    POW = r'\^'
    EQ = r'=='
    GT = r'>'
    GTE = r'>='
    LT = r'<'
    LTE = r'<='
    NE = r'!='
    ARROW = r'='

    # String
    @_(r'''("[^"\\]*(\\.[^"\\]*)*"|'[^'\\]*(\\.[^'\\]*)*')''')
    def STRING(self, t):
        t.value = self.remove_quotes(t.value)
        return t

    def remove_quotes(self, text: str):
        if text.startswith('\"') or text.startswith('\''):
            return text[1:-1]
        return text


    # Extra action for newlines
    def ignore_newline(self, t):
        self.lineno += t.value.count('\n')

    def error(self, t):
        print("Illegal character '%s'" % t.value[0])
        self.index += 1



from setuptools import setup, find_packages
 
classifiers = [
  'Intended Audience :: Developers',
  'Operating System :: OS Independent',
  'License :: OSI Approved :: MIT License',
  'Programming Language :: Python :: 3'
]

setup(
  name='swas',
  version='1.8',
  description='A Progamming Language.',
  long_description=open('README.md').read(),
  long_description_content_type = "text/markdown",
  url = "https://github.com/CodeWithSwastik/swas", 
  project_urls={
   "Documentation": "https://github.com/CodeWithSwastik/swas/tree/main#getting-started",
   "Issue tracker": "https://github.com/CodeWithSwastik/swas/issues",
   },
  author='Swas.py',
  author_email='cwswas.py@gmail.com',
  license='MIT', 
  classifiers=classifiers,
  keywords='swas,language,swas lang', 
  packages=find_packages(),
  install_requires= open('requirements.txt').readlines(),
  python_requires='>=3.6'
)